@@ -31,6 +31,8 @@ struct AllocateCheckerInfo {
   bool gotTypeSpec{false};
   bool gotSource{false};
   bool gotMold{false};
+  bool gotStream{false};
+  bool gotPinned{false};
 };
 
 class AllocationCheckerHelper {
@@ -141,7 +143,10 @@ static std::optional<AllocateCheckerInfo> CheckAllocateOptions(
                         }
                         info.gotStat = true;
                       },
-                      [&](const parser::MsgVariable &) {
+                      [&](const parser::MsgVariable &var) {
+                        WarnOnDeferredLengthCharacterScalar(context,
+                            GetExpr(context, var),
+                            var.v.thing.thing.GetSource(), "ERRMSG=");
                         if (info.gotMsg) { // C943
                           context.Say(
                               "ERRMSG may not be duplicated in a ALLOCATE statement"_err_en_US);
@@ -179,6 +184,22 @@ static std::optional<AllocateCheckerInfo> CheckAllocateOptions(
               parserSourceExpr = &mold.v.value();
               info.gotMold = true;
             },
+            [&](const parser::AllocOpt::Stream &stream) { // CUDA
+              if (info.gotStream) {
+                context.Say(
+                    "STREAM may not be duplicated in a ALLOCATE statement"_err_en_US);
+                stopCheckingAllocate = true;
+              }
+              info.gotStream = true;
+            },
+            [&](const parser::AllocOpt::Pinned &pinned) { // CUDA
+              if (info.gotPinned) {
+                context.Say(
+                    "PINNED may not be duplicated in a ALLOCATE statement"_err_en_US);
+                stopCheckingAllocate = true;
+              }
+              info.gotPinned = true;
+            },
         },
         allocOpt.u);
   }
@@ -350,30 +371,24 @@ static std::optional<std::int64_t> GetTypeParameterInt64Value(
   if (const ParamValue *
       paramValue{derivedType.FindParameter(parameterSymbol.name())}) {
     return evaluate::ToInt64(paramValue->GetExplicit());
-  } else {
-    return std::nullopt;
   }
+  return std::nullopt;
 }
 
-// HaveCompatibleKindParameters functions assume type1 is type compatible with
-// type2 (except for kind type parameters)
-static bool HaveCompatibleKindParameters(
+static bool HaveCompatibleTypeParameters(
     const DerivedTypeSpec &derivedType1, const DerivedTypeSpec &derivedType2) {
   for (const Symbol &symbol :
       OrderParameterDeclarations(derivedType1.typeSymbol())) {
-    if (symbol.get<TypeParamDetails>().attr() == common::TypeParamAttr::Kind) {
-      // At this point, it should have been ensured that these contain integer
-      // constants, so die if this is not the case.
-      if (GetTypeParameterInt64Value(symbol, derivedType1).value() !=
-          GetTypeParameterInt64Value(symbol, derivedType2).value()) {
-        return false;
-      }
+    auto v1{GetTypeParameterInt64Value(symbol, derivedType1)};
+    auto v2{GetTypeParameterInt64Value(symbol, derivedType2)};
+    if (v1 && v2 && *v1 != *v2) {
+      return false;
     }
   }
   return true;
 }
 
-static bool HaveCompatibleKindParameters(
+static bool HaveCompatibleTypeParameters(
     const DeclTypeSpec &type1, const evaluate::DynamicType &type2) {
   if (type1.category() == DeclTypeSpec::Category::ClassStar) {
     return true;
@@ -383,28 +398,56 @@ static bool HaveCompatibleKindParameters(
   } else if (type2.IsUnlimitedPolymorphic()) {
     return false;
   } else if (const DerivedTypeSpec * derivedType1{type1.AsDerived()}) {
-    return HaveCompatibleKindParameters(
+    return HaveCompatibleTypeParameters(
         *derivedType1, type2.GetDerivedTypeSpec());
   } else {
     common::die("unexpected type1 category");
   }
 }
 
-static bool HaveCompatibleKindParameters(
+static bool HaveCompatibleTypeParameters(
     const DeclTypeSpec &type1, const DeclTypeSpec &type2) {
   if (type1.category() == DeclTypeSpec::Category::ClassStar) {
     return true;
-  }
-  if (const IntrinsicTypeSpec * intrinsicType1{type1.AsIntrinsic()}) {
-    return intrinsicType1->kind() == DEREF(type2.AsIntrinsic()).kind();
+  } else if (const IntrinsicTypeSpec * intrinsicType1{type1.AsIntrinsic()}) {
+    const IntrinsicTypeSpec *intrinsicType2{type2.AsIntrinsic()};
+    return !intrinsicType2 || intrinsicType1->kind() == intrinsicType2->kind();
   } else if (const DerivedTypeSpec * derivedType1{type1.AsDerived()}) {
-    return HaveCompatibleKindParameters(
-        *derivedType1, DEREF(type2.AsDerived()));
+    const DerivedTypeSpec *derivedType2{type2.AsDerived()};
+    return !derivedType2 ||
+        HaveCompatibleTypeParameters(*derivedType1, *derivedType2);
   } else {
     common::die("unexpected type1 category");
   }
 }
 
+static bool HaveCompatibleLengths(
+    const DeclTypeSpec &type1, const DeclTypeSpec &type2) {
+  if (type1.category() == DeclTypeSpec::Character &&
+      type2.category() == DeclTypeSpec::Character) {
+    auto v1{
+        evaluate::ToInt64(type1.characterTypeSpec().length().GetExplicit())};
+    auto v2{
+        evaluate::ToInt64(type2.characterTypeSpec().length().GetExplicit())};
+    return !v1 || !v2 || *v1 == *v2;
+  } else {
+    return true;
+  }
+}
+
+static bool HaveCompatibleLengths(
+    const DeclTypeSpec &type1, const evaluate::DynamicType &type2) {
+  if (type1.category() == DeclTypeSpec::Character &&
+      type2.category() == TypeCategory::Character) {
+    auto v1{
+        evaluate::ToInt64(type1.characterTypeSpec().length().GetExplicit())};
+    auto v2{type2.knownLength()};
+    return !v1 || !v2 || *v1 == *v2;
+  } else {
+    return true;
+  }
+}
+
 bool AllocationCheckerHelper::RunChecks(SemanticsContext &context) {
   if (!symbol_) {
     CHECK(context.AnyFatalError());
@@ -455,10 +498,15 @@ bool AllocationCheckerHelper::RunChecks(SemanticsContext &context) {
           "Allocatable object in ALLOCATE must be type compatible with type-spec"_err_en_US);
       return false;
     }
-    if (!HaveCompatibleKindParameters(*type_, *allocateInfo_.typeSpec)) {
+    if (!HaveCompatibleTypeParameters(*type_, *allocateInfo_.typeSpec)) {
       context.Say(name_.source,
           // C936
-          "Kind type parameters of allocatable object in ALLOCATE must be the same as the corresponding ones in type-spec"_err_en_US);
+          "Type parameters of allocatable object in ALLOCATE must be the same as the corresponding ones in type-spec"_err_en_US);
+      return false;
+    }
+    if (!HaveCompatibleLengths(*type_, *allocateInfo_.typeSpec)) { // C934
+      context.Say(name_.source,
+          "Character length of allocatable object in ALLOCATE must be the same as the type-spec"_err_en_US);
       return false;
     }
     if (!HaveSameAssumedTypeParameters(*type_, *allocateInfo_.typeSpec)) {
@@ -474,11 +522,18 @@ bool AllocationCheckerHelper::RunChecks(SemanticsContext &context) {
           "Allocatable object in ALLOCATE must be type compatible with source expression from MOLD or SOURCE"_err_en_US);
       return false;
     }
-    if (!HaveCompatibleKindParameters(
+    if (!HaveCompatibleTypeParameters(
             *type_, allocateInfo_.sourceExprType.value())) {
       // C946
       context.Say(name_.source,
-          "Kind type parameters of allocatable object must be the same as the corresponding ones of SOURCE or MOLD expression"_err_en_US);
+          "Derived type parameters of allocatable object must be the same as the corresponding ones of SOURCE or MOLD expression"_err_en_US);
+      return false;
+    }
+    // Character length distinction is allowed, with a warning
+    if (!HaveCompatibleLengths(
+            *type_, allocateInfo_.sourceExprType.value())) { // C945
+      context.Say(name_.source,
+          "Character length of allocatable object in ALLOCATE should be the same as the SOURCE or MOLD"_port_en_US);
       return false;
     }
   }
@@ -533,12 +588,13 @@ bool AllocationCheckerHelper::RunChecks(SemanticsContext &context) {
     return false;
   }
   context.CheckIndexVarRedefine(name_);
+  const Scope &subpScope{
+      GetProgramUnitContaining(context.FindScope(name_.source))};
   if (allocateObject_.typedExpr && allocateObject_.typedExpr->v) {
-    if (auto whyNot{
-            WhyNotDefinable(name_.source, context.FindScope(name_.source),
-                {DefinabilityFlag::PointerDefinition,
-                    DefinabilityFlag::AcceptAllocatable},
-                *allocateObject_.typedExpr->v)}) {
+    if (auto whyNot{WhyNotDefinable(name_.source, subpScope,
+            {DefinabilityFlag::PointerDefinition,
+                DefinabilityFlag::AcceptAllocatable},
+            *allocateObject_.typedExpr->v)}) {
       context
           .Say(name_.source,
               "Name in ALLOCATE statement is not definable"_err_en_US)
