@@ -69,3 +69,15 @@ subroutine ubound_test_3(a, dim, res)
 ! CHECK:         fir.store %[[VAL_16]] to %{{.*}} : !fir.ref<i64>
   res = ubound(a, dim, 8)
 end subroutine
+
+
+! CHECK-LABEL: func @_QPubound_test_const_dim(
+subroutine ubound_test_const_dim(array)
+  real :: array(11:)
+  integer :: res
+! Should not call _FortranASizeDim when dim is compile time constant. But instead load from descriptor directly.
+! CHECK:         %[[C0:.*]] = arith.constant 0 : index
+! CHECK:         %[[DIMS:.*]]:3 = fir.box_dims %arg0, %[[C0]] : (!fir.box<!fir.array<?xf32>>, index) -> (index, index, index)
+! CHECK:         %{{.*}} = fir.convert %[[DIMS]]#1 : (index) -> i32
+  res = ubound(array, 1)
+end subroutine
